(* Unterroutinen fr Wetter-Programme

    2007-2024 J. Rathlev 24222 Schwentinental
      Web:  www.rathlev-home.de
      Mail: kontakt(a)rathlev-home.de

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   letzte nderung: Feb. 2024
   *)

unit WetSub;

interface

uses System.SysUtils, Vcl.Dialogs, StringUtils;

const
  MTage : array[1..12] of word = (31,28,31,30,31,30,31,31,30,31,30,31);

  DecSep = '.';   // fr Daten in Datei
  SnowId = 'S';

  TNoVal = 99.0;
  DNoval = 0.0;
  RNoval = -1.0;

  WDatName = 'W';
  WDatExt  = '.WDT';
  DatDelim = ',';

  Ws0Id = 'WS data file';
  Ws4Id = 'WS4data file';
  IdLength = length(Ws0Id);

type
  TDatRecTyp = record
    TMin,T,TMax,Regen,Druck : double;
    Schnee                  : boolean;
    Komm                    : string[29];
    end;

  TDatBufTyp = array [1..12,1..31] of TDatRecTyp;

function Valid (x,NoVal : double) : boolean;

(* Anzahl der Tage im Monat *)
function MonatsTage (monat,jahr : word) : word;

(* Daten lesen *)
function ReadData (DatName : string;
                   Jahr    : word;
                   var Buffer : TDatBufTyp) : boolean;

// letzten Tag des Vorjahres laden
procedure ReadLastPreviousDay (DatPath: string;
                               Jahr     : word;
                               var Data : TDatRecTyp);

// ersten Tag des nchsten Jahres laden
procedure ReadFirstNextDay (DatPath : string;
                            Jahr     : word;
                            var Data : TDatRecTyp);

(* Daten schreiben *)
procedure WriteData (DatName : string;
                     Jahr    : word;
                     const Buffer : TDatBufTyp);

(* Dateinamen erzeugen *)
function CreateFileName (Jahr : word) : string;

(* Datensatz initialisieren *)
procedure InitDaten (var Satz : TDatRecTyp);

(* Datenbuffer lschen *)
procedure ClearBuffer (Jahr       : word;
                       var Buffer : TDatBufTyp);

implementation

uses System.StrUtils;

{ ---------------------------------------------------------------- }
(* Wert auf Gltigkeit prfen
   true : x <> NoVal *)
function Valid (x,NoVal : double) : boolean;
begin
  Valid:=abs(x-NoVal)>0.1;
  end;

{-----------------------------------------------------------------------}
(* Anzahl der Tage im Monat *)
function MonatsTage (monat,jahr : word) : word;
var
  i : word;
begin
  i:=mtage[monat];
  if (jahr mod 4 = 0) and (monat=2) then i:=29;
  MonatsTage:=i;
  end;

{-----------------------------------------------------------------------}
(* Daten lesen *)
function ReadData (DatName : string;
                   Jahr    : word;
                   var Buffer : TDatBufTyp) : boolean;
var
  fInp     : TextFile;
  s,sr     : string;
  m,tg     : word;
  NeuMonat,
  ende,err : boolean;

  function ReadNxtVal (var s   : String;
                       Del     : char;
                       Default : double;
                       var err : boolean;
                       DecSep  : char = #0) : double;
  var
    i    : integer;
    x    : double;
    fs   : TFormatSettings;
    t    : string;
  begin
    fs:=FormatSettings;
    if DecSep<>#0 then fs.DecimalSeparator:=DecSep;
    s:=TrimLeft(s); i:=pos(Del,s);
    if i=0 then i:=succ(length(s));
    t:=Trim(copy(s,1,pred(i)));
    if length(t)=0 then Result:=Default
    else if TryStrToFloat(t,x,fs) then Result:=x
    else begin
      Result:=Default;
      err:=true;
      end;
    delete(s,1,i);
    end;

  function GetVal (const s : string;
                   Default : double;
                   var err : boolean;
                   DecSep  : char = #0) : double;
  var
    fs   : TFormatSettings;
  begin
    fs:=FormatSettings;
    if DecSep<>#0 then fs.DecimalSeparator:=DecSep;
    if not TryStrToFloat(Trim(s),Result,fs) then begin
      Result:=Default;
      err:=true;
      end;
    end;

begin
  AssignFile (fInp,DatName);
  try
    Reset (fInp);
    repeat
      readln (fInp,s); ende:=Eof(fInp);
      until AnsiStartsText('#',s) or Ende;
    if not Ende then begin
      ClearBuffer (Jahr,Buffer);
      NeuMonat:=false;
      repeat
        err:=false;
        delete (s,1,1);
        m:=ReadNxtInt(s,DatDelim,0,err);
        if (m>=1) and (m<=12) then repeat (* Monat einlesen *)
          ende:=Eof(fInp);
          if not Ende then begin
            readln (fInp,s); NeuMonat:=AnsiStartsText('#',s);
            if (length(s)>0) and not NeuMonat then begin
              tg:=ReadNxtInt (s,DatDelim,0,err);
              if (tg>=1) and (tg<=31) then with Buffer[m,tg] do begin
                tmin:=ReadNxtVal(s,DatDelim,99.0,err,DecSep);
                t:=ReadNxtVal(s,DatDelim,99.0,err,DecSep);
                tmax:=ReadNxtVal(s,DatDelim,99.0,err,DecSep);
                sr:=Trim(ReadNxtStr(s,DatDelim,'-1.0'));
                if sr[1]=SnowId then begin
                  schnee:=true; delete(sr,1,1);
                  end
                else schnee:=false;
                regen:=GetVal(sr,-1.0,err,DecSep);
//                ReadNxtV(s,DatDelim,-1.0,err,DecSep);
                druck:=ReadNxtVal(s,DatDelim,0.0,err,DecSep);
                Komm:=s;
                end;
              end;
            end;
          until NeuMonat or Ende
        else repeat
          readln (fInp,s); ende:=Eof(fInp)
          until (s[1]='#') or Ende;
        if err and (m>0)then
          ShowMessage ('Fehler in Daten fr '+IntToStr(m)+'/'+
                        IntToStr(Jahr)+'!');
        until ende;
      end;
    CloseFile (fInp);
    Result:=true;
  except
    Result:=false;
    end;
  end;

// letzten Tag des Vorjahres laden
procedure ReadLastPreviousDay (DatPath : string;
                               Jahr     : word;
                               var Data : TDatRecTyp);
var
  Buffer : TDatBufTyp;
begin
  if ReadData(DatPath+CreateFileName(Jahr-1),Jahr-1,Buffer) then Data:=Buffer[12,31]
  else InitDaten(Data);
  end;

// ersten Tag des nchsten Jahres laden
procedure ReadFirstNextDay (DatPath : string;
                            Jahr     : word;
                            var Data : TDatRecTyp);
var
  Buffer : TDatBufTyp;
begin
  if ReadData(DatPath+CreateFileName(Jahr+1),Jahr+1,Buffer) then Data:=Buffer[1,1]
  else InitDaten(Data);
  end;

{-----------------------------------------------------------------------}
(* Daten schreiben *)
procedure WriteData (DatName : string;
                     Jahr    : word;
                     const Buffer : TDatBufTyp);
var
  fOut     : TextFile;
  m,tg     : word;
  sr       : string;
begin
  AssignFile (fOut,DatName);
  Rewrite (fOut);
  for m:=1 to 12 do begin
    writeln (fOut,'#',m,DatDelim,Jahr);
    for tg:=1 to MonatsTage(m,Jahr) do with Buffer[m,tg] do begin
      if schnee then sr:=SnowId else sr:=' ';
      writeln (fOut,tg:2,DatDelim,TMin:5:1,DatDelim,T:5:1,DatDelim,TMax:5:1,
               DatDelim,sr,Regen:5:1,DatDelim,Druck:4:0,DatDelim,Komm);
      end;
    end;
  CloseFile (fOut);
  end;

{-----------------------------------------------------------------------}
(* Dateinamen erzeugen *)
function CreateFileName (Jahr : word) : string;
begin
  CreateFileName:=WDatName+IntToStr(Jahr)+WDatExt;
  end;

{-----------------------------------------------------------------------}
(* Datensatz initialisieren *)
procedure InitDaten (var Satz : TDatRecTyp);
begin
  with Satz do begin
    tmin:=TNoVal; t:=TNoVal; tmax:=TNoVal;
    regen:=RNoVal; schnee:=false; druck:=DNoVal;
    Komm:='';
    end;
  end;

{-----------------------------------------------------------------------}
(* Datenbuffer lschen *)
procedure ClearBuffer (Jahr       : word;
                       var Buffer : TDatBufTyp);
var
  m,t : word;
begin
  for m:=1 to 12 do for t:=1 to MonatsTage(m,Jahr) do
        InitDaten (Buffer[m,t]);
end;

end.
